==> kaffe-0.6.1.tgz: kaffe-0.6.1/kaffevm/exception.c <==
/*
 * exception.c
 * Handle exceptions for the interpreter or translator.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "config-signal.h"
#include <setjmp.h>
#include "gtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "exception.h"
#include "baseClasses.h"
#include "lookup.h"
#include "thread.h"
#include "errors.h"
#include "itypes.h"
#include "md.h"

#if defined(INTERPRETER)
extern vmException* cjbuf;
#define	DEFINEFRAME()		/* Does nothing */
#define	FIRSTFRAME(f, e)	/* Does nothing */
#define	DISPATCHFRAME(e)	dispatchException((throwable*)(e), 0)
#define	EXCEPTIONPROTO		int sig
#define	EXCEPTIONFRAME(f, c)	/* Does nothing */
#define	EXCEPTIONFRAMEPTR	0
#elif defined(TRANSLATOR)
#define	DEFINEFRAME()		exceptionFrame frame
#define	DISPATCHFRAME(e)	dispatchException((throwable*)(e), &frame)
#define	EXCEPTIONFRAMEPTR	&frame
#endif

static void dispatchException(throwable*, struct _exceptionFrame*);
object* buildStackTrace(struct _exceptionFrame*);

extern object* exceptionObject;

/*
 * Throw an internal exception.
 */
void
throwException(object* eobj)
{
	if (eobj != 0) {
		((throwable*)eobj)->backtrace = buildStackTrace(0);
	}
	throwExternalException(eobj);
}

/*
 * Throw an exception.
 */
void
throwExternalException(object* eobj)
{
	DEFINEFRAME();
	if (eobj == 0) {
		fprintf(stderr, "Exception thrown on null object ... aborting\n");
		abort();
		exit(1);
	}
	FIRSTFRAME(frame, eobj);
	DISPATCHFRAME(eobj);
}

static
void
dispatchException(throwable* eobj, struct _exceptionFrame* baseframe)
{
	exceptionInfo info;
	int i;
	object* str;
	char* cname;
	classes* class;

	class = OBJECT_CLASS(&eobj->head);
	/* Search down exception stack for a match */
#if defined(INTERPRETER)
	{
		vmException* frame;
		for (frame = cjbuf; frame != 0; frame = frame->prev) {
			exceptionObject = (object*)eobj;

			/* Look for handler - jump into method if found */
			if (findExceptionBlockInMethod(frame->pc, exceptionObject->dtable->class, frame->meth, &info) == true) {
				cjbuf = frame;
				cjbuf->pc = info.handler;
				longjmp(cjbuf->jbuf, 1);
			}
			/* If not here, exit monitor if synchronised. */
			if (frame->mobj != 0) {
				soft_monitorexit(frame->mobj);
			}
		}
	}
#elif defined(TRANSLATOR)
	{
		exceptionFrame* frame;
		exceptionInfo einfo;
		for (frame = baseframe; FRAMEOKAY(frame); frame = NEXTFRAME(frame)) {
			findExceptionInMethod(PCFRAME(frame), class, &einfo);
			if (einfo.handler != 0) {
				CALL_KAFFE_EXCEPTION(frame, einfo, eobj);
			}
			/* If method found and synchronised, unlock the lock */
			if (einfo.method != 0 && (einfo.method->accflags & ACC_SYNCHRONISED)) {
				if (einfo.method->accflags & ACC_STATIC) {
					soft_monitorexit(&einfo.class->head);
				}
				else {
					soft_monitorexit(FRAMEOBJECT(frame));
				}
			}
		}
	}
#endif

	/* We must catch 'java.lang.ThreadDeath' exceptions now and
	 * kill the thread rather than the machine.
	 */
	cname = CLASS_CNAME(class);
	if (strcmp(cname, THREADDEATHCLASS) == 0) {
		killThread(0);
	}

	fprintf(stderr, "%s", cname);
	/* If we've a detailed message print that too */
	if (eobj->message != 0) {
		cname = makeCString(eobj->message);
		if (cname != 0 && cname[0] != 0) {
			fprintf(stderr, ": %s", cname);
			checked_free(cname);
		}
	}
	fprintf(stderr, "\n");
	if (eobj->backtrace != 0) {
		for (i = 0; i < eobj->backtrace->size; i++) {
			str = ((object**)(eobj->backtrace+1))[i];
			if (str != 0) {
				fprintf(stderr, "%s\n", (char*)(str+1));
			}
		}
	}
	else {
		fprintf(stderr, "\t[no backtrace available]\n");
	}
	exit(1);
}

/*
 * Build an array of char[] for the current stack backtrace.
 */
object*
buildStackTrace(struct _exceptionFrame* base)
{
	char buf[100];
	object* str;
	object* strarray;
	int cnt;
	int i;
	int len;
	methods* meth;
	uintp pc;

#if defined(INTERPRETER)
	vmException* frame;

	cnt = 0;
	/* First count the stack frames */
	for (frame = cjbuf; frame != 0; frame = frame->prev) {
		cnt++;
	}
#elif defined(TRANSLATOR)
	exceptionFrame nframe;
	exceptionFrame* frame;

	if (base == 0) { 
		FIRSTFRAME(nframe, base);
		base = &nframe;
	}
	/* First count the stack frames */
	cnt = 0;
	for (frame = base; FRAMEOKAY(frame); frame = NEXTFRAME(frame)) {
		cnt++;
	}
#endif

	/* Build an array of strings */
	strarray = alloc_objectarray(cnt, "[[B");
	assert(strarray != 0);

	cnt = 0;
#if defined(INTERPRETER)
	for (frame = cjbuf; frame; frame = frame->prev) {
		meth = frame->meth;
		pc = frame->pc;
#elif defined(TRANSLATOR)
	for (frame = base; FRAMEOKAY(frame); frame = NEXTFRAME(frame)) {
		meth = findMethodFromPC(PCFRAME(frame));
		pc = PCFRAME(frame);
#endif
		if (meth != 0) {
			sprintf(buf, "\tat %s.%s(0x%x)",
				CLASS_CNAME(meth->class),
				meth->name->data,
				pc);
			len = strlen(buf);
			str = alloc_array(len, TYPE_Byte);
			assert(str != 0);
			memcpy ((char*)(str+1), buf, len);
		}
		else {
			str = 0;
		}
		((object**)(strarray+1))[cnt] = str;
		cnt++;
	}
	return (strarray);
}

/*
 * Null exception - catches bad memory accesses.
 */
void
nullException(EXCEPTIONPROTO)
{
	throwable* npe;
	sigset_t nsig;
	DEFINEFRAME();

#if !defined(DEBUG)
	/* Reset signal handler */
	signal(sig, (SIG_T)nullException);
	sigemptyset(&nsig);
	sigaddset(&nsig, sig);
	sigprocmask(SIG_UNBLOCK, &nsig, 0);     
#endif
	EXCEPTIONFRAME(frame, ctx);
	npe = (throwable*)NullPointerException;
	npe->backtrace = buildStackTrace(EXCEPTIONFRAMEPTR);
	DISPATCHFRAME(npe);
}

/*
 * Division by zero.
 */
void
arithmeticException(EXCEPTIONPROTO)
{
	throwable* ae;
	sigset_t nsig;
	DEFINEFRAME();

#if !defined(DEBUG)
        /* Reset signal handler - necessary for SysV, does no harm for BSD */
        signal(sig, (SIG_T)arithmeticException);
	sigemptyset(&nsig);
	sigaddset(&nsig, sig);
	sigprocmask(SIG_UNBLOCK, &nsig, 0);     
#endif
	EXCEPTIONFRAME(frame, ctx);
	ae = (throwable*)ArithmeticException;
	ae->backtrace = buildStackTrace(EXCEPTIONFRAMEPTR);
	DISPATCHFRAME(ae);
}

==> kaffe-0.7.0.tgz: kaffe-0.7.0/kaffevm/exception.c <==
/*
 * exception.c
 * Handle exceptions for the interpreter or translator.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "config-signal.h"
#include <setjmp.h>
#include "gtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "exception.h"
#include "baseClasses.h"
#include "lookup.h"
#include "thread.h"
#include "errors.h"
#include "itypes.h"
#include "md.h"

#if defined(INTERPRETER)
extern vmException* cjbuf;
#define	DEFINEFRAME()		/* Does nothing */
#define	FIRSTFRAME(f, e)	/* Does nothing */
#define	DISPATCHFRAME(e)	dispatchException((throwable*)(e), 0)
#define	EXCEPTIONPROTO		int sig
#define	EXCEPTIONFRAME(f, c)	/* Does nothing */
#define	EXCEPTIONFRAMEPTR	0
#elif defined(TRANSLATOR)
#define	DEFINEFRAME()		exceptionFrame frame
#define	DISPATCHFRAME(e)	dispatchException((throwable*)(e), &frame)
#define	EXCEPTIONFRAMEPTR	&frame
#endif

static void dispatchException(throwable*, struct _exceptionFrame*);
object* buildStackTrace(struct _exceptionFrame*);

extern object* exceptionObject;

/*
 * Throw an internal exception.
 */
void
throwException(object* eobj)
{
	if (eobj != 0) {
		((throwable*)eobj)->backtrace = buildStackTrace(0);
	}
	throwExternalException(eobj);
}

/*
 * Throw an exception.
 */
void
throwExternalException(object* eobj)
{
	DEFINEFRAME();
	if (eobj == 0) {
		fprintf(stderr, "Exception thrown on null object ... aborting\n");
		abort();
		exit(1);
	}
	FIRSTFRAME(frame, eobj);
	DISPATCHFRAME(eobj);
}

static
void
dispatchException(throwable* eobj, struct _exceptionFrame* baseframe)
{
	exceptionInfo info;
	int i;
	object* str;
	char* cname;
	classes* class;

	class = OBJECT_CLASS(&eobj->head);
	/* Search down exception stack for a match */
#if defined(INTERPRETER)
	{
		vmException* frame;
		for (frame = cjbuf; frame != 0; frame = frame->prev) {
			exceptionObject = (object*)eobj;

			/* Look for handler - jump into method if found */
			if (findExceptionBlockInMethod(frame->pc, exceptionObject->dtable->class, frame->meth, &info) == true) {
				cjbuf = frame;
				cjbuf->pc = info.handler;
				longjmp(cjbuf->jbuf, 1);
			}
			/* If not here, exit monitor if synchronised. */
			if (frame->mobj != 0) {
				soft_monitorexit(frame->mobj);
			}
		}
	}
#elif defined(TRANSLATOR)
	{
		exceptionFrame* frame;
		exceptionInfo einfo;
		for (frame = baseframe; FRAMEOKAY(frame); NEXTFRAME(frame)) {
			findExceptionInMethod(PCFRAME(frame), class, &einfo);
			if (einfo.handler != 0) {
				CALL_KAFFE_EXCEPTION(frame, einfo, eobj);
			}
			/* If method found and synchronised, unlock the lock */
			if (einfo.method != 0 && (einfo.method->accflags & ACC_SYNCHRONISED)) {
				if (einfo.method->accflags & ACC_STATIC) {
					soft_monitorexit(&einfo.class->head);
				}
				else {
					soft_monitorexit(FRAMEOBJECT(frame));
				}
			}
		}
	}
#endif

	/* We must catch 'java.lang.ThreadDeath' exceptions now and
	 * kill the thread rather than the machine.
	 */
	cname = CLASS_CNAME(class);
	if (strcmp(cname, THREADDEATHCLASS) == 0) {
		killThread(0);
	}

	fprintf(stderr, "%s", cname);
	/* If we've a detailed message print that too */
	if (eobj->message != 0) {
		cname = makeCString(eobj->message);
		if (cname != 0 && cname[0] != 0) {
			fprintf(stderr, ": %s", cname);
			checked_free(cname);
		}
	}
	fprintf(stderr, "\n");
	if (eobj->backtrace != 0) {
		for (i = 0; i < eobj->backtrace->size; i++) {
			str = ((object**)(eobj->backtrace+1))[i];
			if (str != 0) {
				fprintf(stderr, "%s\n", (char*)(str+1));
			}
		}
	}
	else {
		fprintf(stderr, "\t[no backtrace available]\n");
	}
	exit(1);
}

/*
 * Build an array of char[] for the current stack backtrace.
 */
object*
buildStackTrace(struct _exceptionFrame* base)
{
	char buf[100];
	object* str;
	object* strarray;
	int cnt;
	int i;
	int len;
	methods* meth;
	uintp pc;

#if defined(INTERPRETER)
	vmException* frame;

	cnt = 0;
	/* First count the stack frames */
	for (frame = cjbuf; frame != 0; frame = frame->prev) {
		cnt++;
	}
#elif defined(TRANSLATOR)
	exceptionFrame nframe;
	exceptionFrame* frame;

	if (base == 0) { 
		FIRSTFRAME(nframe, base);
		base = &nframe;
	}
	/* First count the stack frames */
	cnt = 0;
	for (frame = base; FRAMEOKAY(frame); NEXTFRAME(frame)) {
		cnt++;
	}
#endif

	/* Build an array of strings */
	strarray = alloc_objectarray(cnt, "[[B");
	assert(strarray != 0);

	cnt = 0;
#if defined(INTERPRETER)
	for (frame = cjbuf; frame; frame = frame->prev) {
		meth = frame->meth;
		pc = frame->pc;
#elif defined(TRANSLATOR)
	for (frame = base; FRAMEOKAY(frame); NEXTFRAME(frame)) {
		meth = findMethodFromPC(PCFRAME(frame));
		pc = PCFRAME(frame);
#endif
		if (meth != 0) {
			sprintf(buf, "\tat %s.%s(0x%x)",
				CLASS_CNAME(meth->class),
				meth->name->data,
				pc);
			len = strlen(buf);
			str = alloc_array(len, TYPE_Byte);
			assert(str != 0);
			memcpy ((char*)(str+1), buf, len);
		}
		else {
			str = 0;
		}
		((object**)(strarray+1))[cnt] = str;
		cnt++;
	}
	return (strarray);
}

/*
 * Null exception - catches bad memory accesses.
 */
void
nullException(EXCEPTIONPROTO)
{
	throwable* npe;
	sigset_t nsig;
	DEFINEFRAME();

#if !defined(DEBUG)
	/* Reset signal handler */
	signal(sig, (SIG_T)nullException);
	sigemptyset(&nsig);
	sigaddset(&nsig, sig);
	sigprocmask(SIG_UNBLOCK, &nsig, 0);     
#endif
	EXCEPTIONFRAME(frame, ctx);
	npe = (throwable*)NullPointerException;
	npe->backtrace = buildStackTrace(EXCEPTIONFRAMEPTR);
	DISPATCHFRAME(npe);
}

/*
 * Division by zero.
 */
void
arithmeticException(EXCEPTIONPROTO)
{
	throwable* ae;
	sigset_t nsig;
	DEFINEFRAME();

#if !defined(DEBUG)
        /* Reset signal handler - necessary for SysV, does no harm for BSD */
        signal(sig, (SIG_T)arithmeticException);
	sigemptyset(&nsig);
	sigaddset(&nsig, sig);
	sigprocmask(SIG_UNBLOCK, &nsig, 0);     
#endif
	EXCEPTIONFRAME(frame, ctx);
	ae = (throwable*)ArithmeticException;
	ae->backtrace = buildStackTrace(EXCEPTIONFRAMEPTR);
	DISPATCHFRAME(ae);
}
